@@ -4,7 +4,7 @@ Miscellaneous utility functions -- anything that doesn't fit into
 one of the other *util.py modules.
 """
 
-__revision__ = "$Id: util.py 86237 2010-11-06 02:58:56Z eric.araujo $"
+__revision__ = "$Id: util.py 86248 2010-11-06 06:00:54Z eric.araujo $"
 
 import sys, os, string, re
 from distutils.errors import DistutilsPlatformError
@@ -235,15 +235,6 @@ def change_root (new_root, pathname):
             path = path[1:]
         return os.path.join(new_root, path)
 
-    elif os.name == 'mac':
-        if not os.path.isabs(pathname):
-            return os.path.join(new_root, pathname)
-        else:
-            # Chop off volume name from start of path
-            elements = pathname.split(":", 1)
-            pathname = ":" + elements[1]
-            return os.path.join(new_root, pathname)
-
     else:
         raise DistutilsPlatformError("nothing known about platform '%s'" % os.name)
 
